from pathlib import Path

import pytest
from load_atoms.dataset_info import DatasetInfo
from pydantic import ValidationError


def get_correct_dictionary():
    """
    get a dictionary that passes validation
    """

    # this file is root/tests/shared/test_dataset_info.py
    project_root = Path(__file__).parent.parent
    correct_yaml_file = project_root / "database" / "C-GAP-17" / "C-GAP-17.yaml"

    return DatasetInfo.from_yaml_file(correct_yaml_file).model_dump()


def test_incorrect_name():
    kwargs = get_correct_dictionary()
    kwargs["name"] = dict(_not="a string")

    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)


def test_incorrect_license():
    kwargs = get_correct_dictionary()
    kwargs["license"] = "Not a real license"

    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)


def test_incorrect_files():
    kwargs = get_correct_dictionary()

    # files should be a mapping of str to checksum

    # not a mapping
    kwargs["files"] = 1
    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)

    # not a mapping of str to checksum
    kwargs["files"] = {1: 1}
    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)

    kwargs["files"] = {"file1": "not a hash"}
    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)


def test_incorrect_citation():
    kwargs = get_correct_dictionary()

    # citation should be a bibtex string
    kwargs["citation"] = "this is not a bibtex string"
    with pytest.raises(ValidationError):
        DatasetInfo(**kwargs)



from pathlib import Path
from typing import Dict, Optional

import yaml
from pydantic import BaseModel, field_validator

from load_atoms.utils import BASE_REMOTE_URL, valid_checksum

valid_licenses = {
    "CC BY-NC-SA 4.0": "https://creativecommons.org/licenses/by-nc-sa/4.0/deed.en",
    "CC BY-NC 4.0": "https://creativecommons.org/licenses/by-nc/4.0/deed.en",
    "CC BY 4.0": "https://creativecommons.org/licenses/by/4.0/deed.en",
    "MIT": "https://opensource.org/licenses/MIT",
}


class DatasetInfo(BaseModel):
    """
    Represents the metadata of a dataset. Anything information that
    is associated with a dataset (but not with a specific structure)
    should be stored here.

    This subclasses pydantic's BaseModel, which means that it can be
    it will be automatically validated when it is created.
    """

    name: str
    """the name of the dataset"""

    description: str
    """a short description of the dataset"""

    files: Dict[str, str]
    """a dictionary mapping file names to their checksums"""

    citation: Optional[str] = None
    """a BibTeX citation for the dataset"""

    license: Optional[str] = None
    """the license of the dataset"""

    representative_structure: Optional[int] = None
    """a representative structure for visualisation purposes"""

    long_description: Optional[str] = None
    """a longer description of the dataset"""

    per_atom_properties: Optional[dict] = None
    """a mapping of per atom property keys to a description"""

    per_structure_properties: Optional[dict] = None
    """a mapping of per structure property keys to a description"""

    url_root: Optional[str] = None
    """the root url of the dataset"""

    @field_validator("license")
    def validate_license(cls, v):
        if v not in valid_licenses:
            raise ValueError(
                f"Invalid license: {v}. Must be one of {list(valid_licenses)}"
            )
        return v

    @field_validator("files")
    def validate_files(cls, v):
        for data in v.values():
            if not valid_checksum(data):
                raise ValueError(f"Invalid checksum: {data}")
        return v

    @field_validator("citation")
    def validate_citation(cls, v):
        v = v.strip()
        if v.startswith("@") and v.endswith("}"):
            return v
        raise ValueError(f"Invalid BibTeX: {v}")

    @classmethod
    def from_yaml_file(cls, path: Path) -> "DatasetInfo":
        """
        Load dataset metadata from a .yaml description file.
        """
        with open(path) as f:
            data = yaml.safe_load(f)
        try:
            return cls(**data)
        except Exception as e:
            raise ValueError(
                f"Error loading dataset description from {path}"
            ) from e

    def remote_file_locations(self) -> Dict[str, str]:
        """
        Mapping from remote file locations to their checksums.
        """
        base_url = self.url_root or BASE_REMOTE_URL + self.name + "/"
        return {base_url + k: v for k, v in self.files.items()}

    @classmethod
    def description_file_url(cls, dataset_id):
        """Get the URL for a dataset description file."""

        return BASE_REMOTE_URL + f"{dataset_id}/{dataset_id}.yaml"


DatasetId = str



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any, Callable, Iterable, overload

import numpy as np
from ase import Atoms
from ase.io import read
from yaml import dump

from . import backend
from .dataset_info import DatasetInfo
from .utils import LazyMapping, frontend_url, intersect, union


def dataset(
    thing: str | list[Atoms] | Path,
    root: str | Path | None = None,
) -> AtomsDataset:
    """
    Load a dataset by name or from a list of structures.

    Parameters
    ----------
    thing
        A dataset id, a list of structures, or a path to a file.
    root
        The root directory to use when loading a dataset by id. If not
        provided, the default root directory (:code:`~/.load-atoms`)
        will be used.

    Examples
    --------
    The following are all viable ways to load a dataset:

    * by id: :code:`dataset("QM7")`
    * by id, with a custom root directory:
      :code:`dataset("QM7", root="./my-datasets")`
    * from a list of structures: :code:`dataset([Atoms("H2O"), Atoms("H2O2")])`
    * from a file: :code:`dataset("path/to/file.xyz")`
    """

    if isinstance(thing, list) and all(isinstance(s, Atoms) for s in thing):
        # thing is a list of structures
        return AtomsDataset(thing)

    if not isinstance(thing, (Path, str)):
        raise TypeError(
            f"Could not load dataset from {thing}. "
            "Please provide a string, a list of structures, "
            "or a path to a file."
        )

    if Path(thing).exists():
        # thing is a string/path to a file that exists
        # assume it is a file containing structures and load them
        structures = read(Path(thing), index=":")
        return AtomsDataset(structures)  # type: ignore

    if isinstance(thing, Path):
        # thing is a path to a file that does not exist
        raise ValueError(f"The provided path does not exist. ({thing})")

    # assume thing is a dataset ID, and try to load it
    return DescribedDataset.from_id(thing, root)


class AtomsDataset:
    """
    A lightweight wrapper around a list of :code:`ase.Atoms` objects.

    Parameters
    ----------
    structures
        A list of :code:`ase.Atoms` objects.
    """

    def __init__(self, structures: list[Atoms]):
        if len(structures) == 1:
            warnings.warn(
                "Creating a dataset with a single structure. "
                "Typically, datasets contain multiple structures - "
                "did you mean to do this?",
                stacklevel=2,
            )

        self.structures = structures

        keys, loader = _get_info_loader(structures)
        self.info = LazyMapping(keys, loader)
        """A key-value mapping of per-structure properties."""

        keys, loader = _get_arrays_loader(structures)
        self.arrays = LazyMapping(keys, loader)
        """A key-value mapping of per-atom properties."""

    def __len__(self):
        return len(self.structures)

    @overload
    def __getitem__(self, index: int) -> Atoms:
        ...

    @overload
    def __getitem__(self, index: slice) -> AtomsDataset:
        ...

    @overload
    def __getitem__(self, index: np.ndarray) -> AtomsDataset:
        ...

    @overload
    def __getitem__(self, index: Iterable[int]) -> AtomsDataset:
        ...

    def __getitem__(self, index: Any):
        # if the passed index is a slice, return a new Dataset object:
        if isinstance(index, slice):
            return AtomsDataset(self.structures[index])

        # if the index is iterable, return a new Dataset object:
        if hasattr(index, "__iter__"):
            # if the index is a np index, we want to keep the same behaviour
            # (e.g. passing array of indices, or a boolean array)
            if isinstance(index, np.ndarray):
                to_keep = np.arange(len(self))[index]
                return AtomsDataset([self.structures[i] for i in to_keep])

            # some other iterable, e.g. a list of indices
            return AtomsDataset([self.structures[i] for i in index])  # type: ignore

        # otherwise, we assume the index is an integer,
        # and return a single structure
        return self.structures[int(index)]

    def __iter__(self):
        return iter(self.structures)

    def __repr__(self):
        return summarise_dataset(self.structures)

    @property
    def structure_sizes(self):
        """The number of atoms in each structure."""
        return np.array([len(s) for s in self.structures])


def usage_info(dataset: DatasetInfo) -> str:
    info = []
    if dataset.license is not None:
        info.append(
            f"The {dataset.name} dataset is covered by the "
            f"{dataset.license} license."
        )
    if dataset.citation is not None:
        info.append(
            f"Please cite the {dataset.name} dataset "
            "if you use it in your work."
        )

    return "\n".join(info)


class DescribedDataset(AtomsDataset):
    def __init__(
        self,
        structures: list[Atoms],
        description: DatasetInfo,
    ):
        super().__init__(structures)
        self.description = description

    @classmethod
    def from_id(
        cls,
        dataset_id: str,
        root: Path | (str | None) = None,
        verbose: bool = True,
    ) -> AtomsDataset:
        """
        Load a dataset by id.

        Parameters
        ----------
        dataset_id : str
            the id of the dataset to load
        root : Union[Path, str, None], optional
            the root directory to cache the dataset to, by default None
        verbose : bool, optional
            whether to print information about the dataset, by default True
        """

        if root is None:
            root = Path.home() / ".load-atoms"
        root = Path(root)

        all_structures, info = backend.get_structures_for(dataset_id, root)

        if verbose:
            print(usage_info(info))
        return cls(all_structures, info)

    def __repr__(self):
        return summarise_dataset(self.structures, self.description)


def _get_info_loader(
    structures: list[Atoms],
) -> tuple[list[str], Callable[[str], Any]]:
    keys = intersect(structure.info.keys() for structure in structures)

    def loader(key: str):
        return np.array([structure.info[key] for structure in structures])

    return list(keys), loader


def _get_arrays_loader(
    structures: list[Atoms],
) -> tuple[list[str], Callable[[str], Any]]:
    keys = intersect(structure.arrays.keys() for structure in structures)

    def loader(key: str):
        return np.concatenate(
            [structure.arrays[key] for structure in structures]
        )

    return list(keys), loader


def summarise_dataset(
    structures: list[Atoms] | AtomsDataset,
    description: DatasetInfo | None = None,
) -> str:
    name = description.name if description is not None else "Dataset"
    N = len(structures)
    number_atoms = sum([len(structure) for structure in structures])

    per_atom_properties = intersect(
        structure.arrays.keys() for structure in structures
    )
    per_atom_properties -= {"numbers", "positions"}
    per_structure_properties = intersect(
        structure.info.keys() for structure in structures
    )

    species = union(
        structure.get_chemical_symbols() for structure in structures
    )
    species_counts = {
        s: sum(
            [
                structure.get_chemical_symbols().count(s)
                for structure in structures
            ]
        )
        for s in species
    }
    species_counts = {
        k: f"{v / number_atoms:.2%}"
        for k, v in sorted(
            species_counts.items(), key=lambda item: item[1], reverse=True
        )
    }

    if N >= 1000:
        N = f"{N:,}"
    if number_atoms >= 1000:
        number_atoms = f"{number_atoms:,}"

    fields = {
        "structures": N,
        "atoms": number_atoms,
        "species": species_counts,
        "properties": {
            "per atom": "(" + ", ".join(per_atom_properties) + ")",
            "per structure": "(" + ", ".join(per_structure_properties) + ")",
        },
    }

    return dump({name: fields}, sort_keys=False, indent=4)



from __future__ import annotations

import warnings
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import requests
from ase import Atoms
from ase.io import read
from rich.progress import Progress, TaskID, track

from load_atoms.dataset_info import DatasetId, DatasetInfo
from load_atoms.utils import UnknownDatasetException, matches_checksum


def get_structures_for(
    dataset_id: DatasetId, root: Path
) -> tuple[list[Atoms], DatasetInfo]:
    """
    Get the structures comprising the dataset with the given id, either by
    downloading them from the web or by loading them from disk at the given
    path.

    Parameters
    ----------
    dataset_id
        The id of the dataset to load.
    path
        The path to save the structures to.

    Returns
    -------
    List[Atoms]
        The structures comprising the dataset.
    """

    storage = DataStorage(root)
    # if the dataset doesn't already exist, download it from the web,
    # validate it and save it to the given path
    storage.download_missing(dataset_id)

    return storage.load_dataset(dataset_id)


def _download_with_progress(
    url: str, local_path: Path, progress: Progress, task: TaskID
):
    """
    download the file at the given url to the given path, and update the given
    progress bar as the file is downloaded.
    """
    if local_path.is_dir():
        local_path = local_path / Path(url).name
    local_path.parent.mkdir(parents=True, exist_ok=True)

    with requests.get(url, stream=True) as response:
        # raise an exception if the request was not successful
        response.raise_for_status()

        file_size = int(response.headers.get("content-length", 0))
        progress.update(task, total=file_size)
        progress.start_task(task)

        with open(local_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    progress.update(task, advance=len(chunk))


def download(url: str, local_path: Path):
    """
    Download a file from the given url to the given path.

    If path is a file, the file will be downloaded to that path.
    Else, the file will be downloaded to the given path, with the same name as
    the file at the given url.

    Parameters
    ----------
    url
        The url to download the file from.
    local_path
        The path to download the file to.
    """

    with Progress(transient=True) as progress:
        task = progress.add_task(f"Downloading {Path(url).name}", start=False)
        _download_with_progress(url, local_path, progress, task)


def download_all(urls: list[str], directory: Path):
    """
    Download all files from the given urls to the given directory.

    Parameters
    ----------
    urls
        A list of urls to download the files from.
    directory
        The directory to download the files to.
    """

    if len(urls) == 0:
        return

    if len(urls) == 1:
        download(urls[0], directory)
        return

    pool_exectutor = ThreadPoolExecutor(
        max_workers=8, thread_name_prefix="download"
    )
    futures = []
    with Progress(transient=True) as progress, pool_exectutor as pool:
        for url in urls:
            task = progress.add_task(
                f"Downloading {Path(url).name}", start=False
            )
            future = pool.submit(
                _download_with_progress, url, directory, progress, task
            )
            futures.append((future, url))

    failures = []
    for future, url in futures:
        try:
            future.result()
        except Exception:
            failures.append(url)

    if len(failures) > 0:
        raise Exception(f"Failed to download {len(failures)} files: {failures}")

    pool_exectutor.shutdown()


class DataStorage:
    """
    Handle the storage of datasets at a specific path.

    Each dataset is stored in a sub-directory with name <dataset_id>.
    Within this directory, the dataset description is stored in a .yaml
    file, together with the structure files.

    Parameters
    ----------
    home
        The path to store the datasets at.
    """

    def __init__(self, root: Path):
        self.root = root

    def _sub_folder_for(self, dataset_id):
        return self.root / dataset_id

    def download_missing(self, dataset_id: DatasetId):
        """
        Download the dataset with the given id from the web and validate it.

        Parameters
        ----------
        dataset_id
            The id of the dataset to download.
        """

        folder = self._sub_folder_for(dataset_id)
        folder.mkdir(parents=True, exist_ok=True)
        info_file = folder / (dataset_id + ".yaml")

        if not info_file.exists():
            # download the dataset description file
            try:
                download(
                    DatasetInfo.description_file_url(dataset_id),
                    info_file,
                )
            except Exception as e:
                raise UnknownDatasetException(dataset_id) from e

        info = DatasetInfo.from_yaml_file(info_file)

        # download any missing dataset files
        missing_files = {
            file: hash
            for file, hash in info.files.items()
            if not (folder / file).exists()
        }

        remote_file_locations = [
            url
            for url in info.remote_file_locations()
            if Path(url).name in missing_files
        ]

        download_all(remote_file_locations, folder)

        # validate the just-downloaded files
        for file_name, hash in missing_files.items():
            file = folder / file_name
            if not matches_checksum(file, hash):
                warnings.warn(
                    f"Checksum of {file_name} does not match the "
                    "expected value. This means that the downloaded dataset "
                    "may be corrupted.",
                    stacklevel=2,
                )

    def load_dataset(self, dataset_id) -> tuple[list[Atoms], DatasetInfo]:
        """
        Load the dataset with the given id from the given path.

        Parameters
        ----------
        dataset_id
            The id of the dataset to load.

        Returns
        -------
        List[Atoms]
            The structures comprising the dataset.
        """

        folder = self._sub_folder_for(dataset_id)
        info_file = folder / (dataset_id + ".yaml")

        info = DatasetInfo.from_yaml_file(info_file)
        structures = []

        iterator = info.files
        if len(info.files) > 3:
            # show a progress bar if there are many files
            iterator = track(
                info.files,
                description=f"Reading files from disk for {dataset_id}",
                transient=True,
            )

        for file_name in iterator:
            file = folder / file_name
            structures.extend(read(file, index=":"))

        return structures, info



import pytest
from load_atoms.utils import LazyMapping, intersect, lpad, union


def test_intersection():
    assert intersect([[1, 2, 3], [2, 3, 4]]) == {2, 3}
    assert intersect(("hi", "hello")) == {"h"}


def test_union():
    assert union([[1, 2, 3], [2, 3, 4]]) == {1, 2, 3, 4}
    assert union(("hi", "hello")) == {"h", "e", "l", "o", "i"}


def test_pad():
    assert lpad("hi", 2) == " " * 2 + "hi"
    assert lpad("hi\nthere", 2) == "  hi\n  there"


def test_lazy_mapping():
    keys = [1, 2, 3]
    _calls = 0

    def loader(key):
        nonlocal _calls
        _calls += 1
        return key + 1

    mapping = LazyMapping(keys, loader)

    with pytest.raises(KeyError):
        mapping[4]

    assert 1 in mapping

    assert mapping[1] == 2
    assert _calls == 1

    mapping[1]
    assert _calls == 1



from load_atoms.utils import generate_checksum, matches_checksum


def test_matches_contents(tmp_path):
    """
    Test that check_file_contents returns True if the contents of a file
    match the given hash.
    """

    fake_file = tmp_path / "file.txt"
    fake_file.write_text("fake file")

    correct_hash = generate_checksum(fake_file)
    incorrect_hash = "0" * 12

    assert matches_checksum(fake_file, correct_hash)
    assert not matches_checksum(fake_file, incorrect_hash)
    assert not matches_checksum(fake_file, correct_hash[:-1])



import pytest
from load_atoms.backend import get_structures_for
from load_atoms.utils import UnknownDatasetException


def test_get_structures_for(tmp_path):
    """
    Test that get_structures_for returns the correct structures
    """

    structures, info = get_structures_for("C-GAP-17", tmp_path)

    assert len(structures) == 4530, "Incorrect number of structures"
    assert info.name == "C-GAP-17", "Incorrect dataset name"

    folder = tmp_path / "C-GAP-17"
    assert (folder / "C-GAP-17.yaml").exists(), "Dataset description missing"
    assert (folder / "C-GAP-17.extxyz").exists(), "Structure file missing"

    with pytest.raises(UnknownDatasetException):
        get_structures_for("made_up_dataset", tmp_path)
